/* ********************************************************
    addeg.cpp

    Author = Hideyuki Harada
    Nihon Univ. Tani Research Group


    My Algorithm for resolving MaxClique problem
    designed for "g++ 3.3.3" with "LEDA 4.5"
   ******************************************************** */


#include<LEDA/graph.h>
#include<LEDA/list.h>
#include<time.h>
#include<iostream.h>
#include<fstream.h>
#include<string.h>
#include<LEDA/d_array.h>

#define P_VERSION "addeg 0.01"

using namespace leda;

graph G;           // undirected graph
list<node> V;
list<edge> E;


list<node> Qmax;   // node set for MaxClique
list<node> Q;

d_array<node, int> ADDEG;

int Branch_count;  // number of counted SST branch

int upper_count;

void read_graph(int argc, char *argv[]);   // read filename from command line

void initialize(list<node> *R);  // clear Q and put V to R

void show_result(float T, char *argv[]);  // write result

void expand(list<node> R);  // basic algorithm

void DEG_CLQ(void);

void sort_by_degree(list<node> *R);

int cmp_addeg(const node &v, const node &u);

void set_addeg(list<node> *R);

int upper_bound(list<node> R);

int cmp_deg(const node &v, const node &u);

//*************************************************


int main(int argc, char *argv[]){

  float T = used_time();

  if(argc != 2){
    cout << "usage: color [file]" << endl;
    exit(1);
  }

  read_graph(argc, argv);

  cout << "resolving " << argv[1] << " ..." << endl;

  // start procedure
  DEG_CLQ();

  show_result(T, argv);

  cout << "done." << endl;

  return 0;

}

void DEG_CLQ(void){

  list<node> R;

  initialize(&R);
  sort_by_degree(&R);
  set_addeg(&R);

  expand(R);

}

void expand(list<node> R){

  node p, v;
  list<node> Rp;

  while(!R.empty()){
    
    Branch_count++;
    p = R.head();
 
    if(Q.size() + ADDEG[p] > Qmax.size()){
      Q.append(p);

      Rp.clear();
      // Rp = intersect(R, G.adj_nodes(p));
      forall_adj_nodes(v, p){
	if(R.search(v) != NULL && ADDEG[v] > Qmax.size()){
	  Rp.append(v);
	}
      } // R no atamakara
      sort_by_degree(&Rp);

      // uuper bound hozonsite tudukikara siraberu        
      if(!Rp.empty() ){
	if(Q.size() + upper_bound(Rp) > Qmax.size()) 
	  expand(Rp);
      }else{
	if(Q.size() > Qmax.size()){  // apply new MaxClique
	  Qmax.clear();
	  forall(v, Q){
	    Qmax.append(v);
	  }
	}
      }
      Q.remove(p);
    }else{
      return;
    }

    R.remove(p);
  }
  return;

}


void sort_by_degree(list<node> *R){
  
  (*R).sort(*cmp_addeg, 0);

  return;

}

int cmp_addeg(const node &v, const node &u){

  if(ADDEG[v] < ADDEG[u]){
    return 1;
  }

  else
    if(ADDEG[v] == ADDEG[u]){
      return 0;
    }

    else
      return -1;
}

void read_graph(int argc, char *argv[]){
  node v;

  G.read(argv[1]);
  G.make_undirected();
  forall_nodes(v, G){
    if(degree(v) == 0)
      G.del_node(v);
  }
  E = G.all_edges();
  V = G.all_nodes();
  
  return;
}

void initialize(list<node> *R){

  Q.clear();
  (*R) = G.all_nodes();
  Qmax.clear();
  Branch_count = 0;
  upper_count =0;

  return;
}

void show_result(float T, char *argv[]){

  node v;
  ofstream fout;
  time_t t=time(NULL);
  string ofile;

  ofile = ".result";
 
  fout.open(argv[1]+ofile, ios::app);

  fout << asctime(localtime(&t));
  fout << "Program version = " << P_VERSION << endl;  
  fout << "Filename = " << argv[1] << endl;
  fout << " Number of vertexes = " << G.number_of_nodes() << endl;
  fout << " Number of edges = " << G.number_of_edges() << endl;
  fout << "Used time = " << used_time(T) << endl;
  fout << "SST Branch count = " << Branch_count << endl;
  fout << "node count for upper bound = " << upper_count << endl;
  fout << "MAX CLIQUE SIZE = " << Qmax.size() << endl;
  fout << "CLIQUE ELEMENT(S) :" << endl;

  forall(v, Qmax){
    fout << " " << index(v);
  }
  fout << endl;

  fout << endl;

  fout.close();
 
  return;
}


void set_addeg(list<node> *R){
  
  node v;
  int max;

  forall(v, *R){
    max = degree((*R)[(G.adj_nodes(v)).max(cmp_deg)]);
    if(degree(v) < max){
      ADDEG[v] = max + 1;
    }else{
      ADDEG[v] = degree(v) + 1;
    }
  }

  return;
}

int cmp_deg(const node &v, const node &u){

  return(compare(degree(v), degree(u)));

}


int upper_bound(list<node> R){
  int i=0;

  if(!(R.empty())){
    while(i < ADDEG[R[R[i]]] && i < R.size()-1){
      i++;
    }
    upper_count += i;
  }else{
    return 0;
  }

  return i;
}
